# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class ARouteServerError(Exception):
    pass

class ConfigError(ARouteServerError):
    pass

class MissingArgumentError(ARouteServerError):

    def __init__(self, arg):
        ARouteServerError.__init__(self)
        self.arg = arg

    def __str__(self):
        return "Missing argument: {}".format(self.arg)

class MissingFileError(ARouteServerError):

    def __init__(self, path):
        ARouteServerError.__init__(self)
        self.path = path

    def __str__(self):
        return "The file {} does not exist".format(self.path)

class MissingDirError(ARouteServerError):

    def __init__(self, path):
        ARouteServerError.__init__(self)
        self.path = path

    def __str__(self):
        return "The directory {} does not exist".format(self.path)

class CachedObjectsError(ARouteServerError):
    pass

class RPSLToolsError(ARouteServerError):
    pass

class PeeringDBError(ARouteServerError):
    pass

class PeeringDBNoInfoError(ARouteServerError):
    pass

class BuilderError(ARouteServerError):
    pass

class ResourceNotFoundError(ARouteServerError):

    def __init__(self, err):
        ARouteServerError.__init__(self)
        self.err = err

    def __str__(self):
        return "Resource not found: {}".format(self.err)



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

import yaml

from .cfg_base import TestConfigParserBase 
from pierky.arouteserver.config.general import ConfigParserGeneral
from pierky.arouteserver.errors import ConfigError


class TestConfigParserGeneric(TestConfigParserBase):

    FILE_PATH = "config.d/general.yml"
    CONFIG_PARSER_CLASS = ConfigParserGeneral

    VALID_STD_COMMS = ("65535:666", "1:1", "0:0", "1:0", "65535:65535", "rs_as:1")
    VALID_LRG_COMMS = ("65535:666:0", "1:1:1", "0:0:0", "1:0:65535", "4294967295:4294967295:4294967295", "rs_as:1:2")
    VALID_EXT_COMMS = ("rt:1:0", "rt:65535:666", "ro:65535:666", "rt:rs_as:3")

    def test_valid_cfg(self):
        """Generic config parser: valid configuration"""
        self._contains_err()

    def test_unknown_statement(self):
        """Generic config parser: unknown statement"""
        self.cfg["test"] = "test"
        self._contains_err("Unknown statement at 'cfg' level: 'test'.")

    def test_rs_as(self):
        """Generic config parser: rs_as"""
        self.assertEqual(self.cfg["rs_as"], 999)
        for asn in (-1, 0, "test"):
            self.cfg["rs_as"] = asn
            self._contains_err("Error parsing 'rs_as' at 'cfg' level - Invalid ASN: {}".format(str(asn)))

        for asn in (1, 65535, 4294967295): 
            self.cfg["rs_as"] = asn
            self._contains_err()

        self._test_mandatory(self.cfg, "rs_as")

    def test_router_id(self):
        """Generic config parser: router_id"""
        self.assertEqual(self.cfg["router_id"], "192.0.2.2")
        for v in ("1.0.0.1", "10.0.0.1"):
            self.cfg["router_id"] = v
            self._contains_err()
        for v in ("10.0.0.1/24", "fe80::1", "2001:db8::1", "test"):
            self.cfg["router_id"] = v
            self._contains_err("Error parsing 'router_id' at 'cfg' level - Invalid IPv4 address: {}.".format(v))
        self._test_mandatory(self.cfg, "router_id")

    def test_prepend_rs_as(self):
        """Generic config parser: prepend_rs_as"""
        self.assertEqual(self.cfg["prepend_rs_as"], False)
        self._test_bool_val(self.cfg, "prepend_rs_as")
        self._test_mandatory(self.cfg, "prepend_rs_as", has_default=True)

    def test_path_hiding(self):
        """Generic config parser: path_hiding"""
        self._test_bool_val(self.cfg, "path_hiding")
        self._test_optional(self.cfg, "path_hiding")

    def test_passive(self):
        """Generic config parser: passive"""
        self.assertEqual(self.cfg["passive"], True)
        self._test_bool_val(self.cfg, "passive")
        self._test_mandatory(self.cfg, "passive", has_default=True)

    def test_gtsm(self):
        """Generic config parser: gtsm"""
        self.assertEqual(self.cfg["gtsm"], False)
        self._test_bool_val(self.cfg, "gtsm")
        self._test_mandatory(self.cfg, "gtsm", has_default=True)

    def test_add_path(self):
        """Generic config parser: add_path"""
        self.assertEqual(self.cfg["add_path"], False)
        self._test_bool_val(self.cfg, "add_path")
        self._test_mandatory(self.cfg, "add_path", has_default=True)

    def test_next_hop_policy(self):
        """Generic config parser: next_hop_policy"""
        self.assertEqual(self.cfg["filtering"]["next_hop_policy"], "strict")
        self._test_option(self.cfg["filtering"], "next_hop_policy", ("strict", "same-as"))
        self._test_mandatory(self.cfg["filtering"], "next_hop_policy", has_default=True)

    def test_ipv4_pref_len(self):
        """Generic config parser: ipv4_pref_len"""
        self.assertEqual(self.cfg["filtering"]["ipv4_pref_len"]["min"], 8)
        self.assertEqual(self.cfg["filtering"]["ipv4_pref_len"]["max"], 24)
        self.cfg["filtering"]["ipv4_pref_len"]["max"] = 32
        self._test_ip_min_max_len(self.cfg["filtering"], "ipv4_pref_len", "min", 4, (0, 1, 32), (-1, 33, "a"))

        self.cfg["filtering"]["ipv4_pref_len"]["min"] = 0
        self._test_ip_min_max_len(self.cfg["filtering"], "ipv4_pref_len", "max", 4, (0, 1, 32), (-1, 33, "a"))

        for pair in ((0, 32), (31, 32), (32, 32)):
            self.cfg["filtering"]["ipv4_pref_len"]["min"] = pair[0]
            self.cfg["filtering"]["ipv4_pref_len"]["max"] = pair[1]
            self._contains_err()
        for pair in ((10, 5), (32, 0)):
            self.cfg["filtering"]["ipv4_pref_len"]["min"] = pair[0]
            self.cfg["filtering"]["ipv4_pref_len"]["max"] = pair[1]
            self._contains_err("Error parsing 'ipv4_pref_len' at 'cfg.filtering' level - In the IPv4 min/max length, the value of 'min' must be <= the value of 'max'.")

    def test_ipv6_pref_len(self):
        """Generic config parser: ipv6_pref_len"""
        self.assertEqual(self.cfg["filtering"]["ipv6_pref_len"]["min"], 12)
        self.assertEqual(self.cfg["filtering"]["ipv6_pref_len"]["max"], 48)
        self.cfg["filtering"]["ipv6_pref_len"]["max"] = 128
        self._test_ip_min_max_len(self.cfg["filtering"], "ipv6_pref_len", "min", 6, (0, 1, 32, 64, 128), (-1, 129, "a"))

        self.cfg["filtering"]["ipv6_pref_len"]["min"] = 0
        self._test_ip_min_max_len(self.cfg["filtering"], "ipv6_pref_len", "max", 6, (0, 1, 32, 64, 128), (-1, 129, "a"))

        for pair in ((0, 64), (32, 48), (32, 32), (128, 128)):
            self.cfg["filtering"]["ipv6_pref_len"]["min"] = pair[0]
            self.cfg["filtering"]["ipv6_pref_len"]["max"] = pair[1]
            self._contains_err()
        for pair in ((10, 5), (32, 0), (128, 0), (128, 64)):
            self.cfg["filtering"]["ipv6_pref_len"]["min"] = pair[0]
            self.cfg["filtering"]["ipv6_pref_len"]["max"] = pair[1]
            self._contains_err("Error parsing 'ipv6_pref_len' at 'cfg.filtering' level - In the IPv6 min/max length, the value of 'min' must be <= the value of 'max'.")

    def test_global_black_list_pref(self):
        """Generic config parser: global_black_list_pref"""

        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2\n"
            "  filtering:\n"
            "    global_black_list_pref:\n"
            "      - prefix: 192.0.2.0\n"
            "        length: 24\n"
            "        comment: 'Local network'\n"
        )
        self.cfg.parse()
        self._contains_err()

        self.assertEqual(self.cfg["filtering"]["global_black_list_pref"][0],
                         {
                             "prefix": "192.0.2.0",
                             "length": 24,
                             "comment": "Local network",
                             "exact": False,
                             "ge": None,
                             "le": None,
                             "max_length": 32
                         })
        self._test_optional(self.cfg["filtering"], "global_black_list_pref")

    def test_max_as_path_len(self):
        """Generic config parser: max_as_path_len"""
        self.assertEqual(self.cfg["filtering"]["max_as_path_len"], 32)
        for v in (0, 65):
            self.cfg["filtering"]["max_as_path_len"] = v
            self._contains_err("Error parsing 'max_as_path_len' at 'cfg.filtering' level - Invalid max_as_path_len: must be an integer between 1 and 64.")
        for v in (1, 64, 32):
            self.cfg["filtering"]["max_as_path_len"] = v
            self._contains_err()
        self._test_mandatory(self.cfg["filtering"], "max_as_path_len", has_default=True)

    def test_reject_invalid_as_in_as_path(self):
        """Generic config parser: reject_invalid_as_in_as_path"""
        self.assertEqual(self.cfg["filtering"]["reject_invalid_as_in_as_path"], True)
        self._test_bool_val(self.cfg["filtering"], "reject_invalid_as_in_as_path")
        self._test_mandatory(self.cfg["filtering"], "reject_invalid_as_in_as_path", has_default=True)

    def test_tag_as_set(self):
        """Generic config parser: tag_as_set"""
        self.assertEqual(self.cfg["filtering"]["rpsl"]["tag_as_set"], True)
        self._test_bool_val(self.cfg["filtering"]["rpsl"], "tag_as_set")
        self._test_mandatory(self.cfg["filtering"]["rpsl"], "tag_as_set", has_default=True)

    def test_enforce_origin_in_as_set(self):
        """Generic config parser: enforce_origin_in_as_set"""
        self.assertEqual(self.cfg["filtering"]["rpsl"]["enforce_origin_in_as_set"], True)
        self._test_bool_val(self.cfg["filtering"]["rpsl"], "enforce_origin_in_as_set")
        self._test_mandatory(self.cfg["filtering"]["rpsl"], "enforce_origin_in_as_set", has_default=True)

    def test_enforce_prefix_in_as_set(self):
        """Generic config parser: enforce_prefix_in_as_set"""
        self.assertEqual(self.cfg["filtering"]["rpsl"]["enforce_prefix_in_as_set"], True)
        self._test_bool_val(self.cfg["filtering"]["rpsl"], "enforce_prefix_in_as_set")
        self._test_mandatory(self.cfg["filtering"]["rpsl"], "enforce_prefix_in_as_set", has_default=True)

    def test_rpki_enabled(self):
        """Generic config parser: rpki, enabled"""
        self.assertEqual(self.cfg["filtering"]["rpki"]["enabled"], True)
        self._test_bool_val(self.cfg["filtering"]["rpki"], "enabled")
        self._test_mandatory(self.cfg["filtering"]["rpki"], "enabled", has_default=True)

    def test_rpki_data_source(self):
        """Generic config parser: rpki, data_source"""
        self.assertEqual(self.cfg["filtering"]["rpki"]["data_source"], "rtrsub")
        self._test_option(self.cfg["filtering"]["rpki"], "data_source", ("rtrsub", "rtrlib"))
        self._test_optional(self.cfg["filtering"]["rpki"], "data_source")

    def test_rpki_reject_invalid(self):
        """Generic config parser: rpki, reject_invalid"""
        self.assertEqual(self.cfg["filtering"]["rpki"]["reject_invalid"], True)
        self._test_bool_val(self.cfg["filtering"]["rpki"], "reject_invalid")
        self._test_optional(self.cfg["filtering"]["rpki"], "reject_invalid")

    def test_blackhole_announce_to_client(self):
        """Generic config parser: blackhole_filtering, announce_to_client"""
        self.assertEqual(self.cfg["blackhole_filtering"]["announce_to_client"], True)
        self._test_bool_val(self.cfg["blackhole_filtering"], "announce_to_client")

    def test_blackhole_filtering_policy_ipv4(self):
        """Generic config parser: blackhole_filtering, policy_ipv4"""
        self.load_config(file_name="{}/test_cfg_general_blackhole_filtering.yml".format(os.path.dirname(__file__)))
        self.assertEqual(self.cfg["blackhole_filtering"]["policy_ipv4"], "propagate-unchanged")

        self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv4"] = "192.0.2.1"
        self._test_option(self.cfg["blackhole_filtering"], "policy_ipv4", ("propagate-unchanged", "rewrite-next-hop"))
        self._test_optional(self.cfg["blackhole_filtering"], "policy_ipv4")

        self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv4"] = ""
        self.cfg["blackhole_filtering"]["policy_ipv4"] = "propagate-unchanged"
        self._contains_err()
        self.cfg["blackhole_filtering"]["policy_ipv4"] = "rewrite-next-hop"
        self._contains_err("Since blackhole_filtering.policy_ipv4 is 'rewrite_next_hop', an IPv4 address must be provided in 'rewrite_next_hop_ipv4'.")

    def test_blackhole_filtering_policy_ipv6(self):
        """Generic config parser: blackhole_filtering, policy_ipv6"""
        self.load_config(file_name="{}/test_cfg_general_blackhole_filtering.yml".format(os.path.dirname(__file__)))
        self.assertEqual(self.cfg["blackhole_filtering"]["policy_ipv6"], "propagate-unchanged")

        self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv6"] = "fe80::1"
        self._test_option(self.cfg["blackhole_filtering"], "policy_ipv6", ("propagate-unchanged", "rewrite-next-hop"))
        self._test_optional(self.cfg["blackhole_filtering"], "policy_ipv6")

        self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv6"] = ""
        self.cfg["blackhole_filtering"]["policy_ipv6"] = "propagate-unchanged"
        self._contains_err()
        self.cfg["blackhole_filtering"]["policy_ipv6"] = "rewrite-next-hop"
        self._contains_err("Since blackhole_filtering.policy_ipv6 is 'rewrite_next_hop', an IPv6 address must be provided in 'rewrite_next_hop_ipv6'.")

    def test_blackhole_filtering_ipv4(self):
        """Generic config parser: blackhole_filtering, rewrite_next_hop, ipv4"""
        self.load_config(file_name="{}/test_cfg_general_blackhole_filtering.yml".format(os.path.dirname(__file__)))
        for ip in ("127.0.0.1", "192.168.0.1", "192.0.2.1", "12.34.56.78"):
            self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv4"] = ip
            self._contains_err()
        for ip in ("10.0.0.1/24", "a", "fe80::1"):
            self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv4"] = ip
            self._contains_err("Error parsing 'rewrite_next_hop_ipv4' at 'cfg.blackhole_filtering' level - Invalid IPv4 address: {}.".format(ip))
        self._test_optional(self.cfg["blackhole_filtering"], "rewrite_next_hop_ipv4")

    def test_blackhole_filtering_ipv6(self):
        """Generic config parser: blackhole_filtering, rewrite_next_hop, ipv6"""
        self.load_config(file_name="{}/test_cfg_general_blackhole_filtering.yml".format(os.path.dirname(__file__)))
        for ip in ("fe80::1", "2001:DB8::1"):
            self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv6"] = ip
            self._contains_err()
        for ip in ("10.0.0.1/24", "a"):
            self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv6"] = ip
            self._contains_err("Error parsing 'rewrite_next_hop_ipv6' at 'cfg.blackhole_filtering' level - Invalid IPv6 address: {}.".format(ip))
        self._test_optional(self.cfg["blackhole_filtering"], "rewrite_next_hop_ipv6")

    def test_control_communities(self):
        """Generic config parser: control_communities"""
        self.assertEqual(self.cfg["control_communities"], True)
        self._test_bool_val(self.cfg, "control_communities")
        self._test_mandatory(self.cfg, "control_communities", has_default=True)

    def test_communities_std(self):
        """Generic config parser: standard BGP communities"""


        self.cfg["communities"]["blackholing"] = {}
        for c in self.VALID_STD_COMMS:
            self.cfg["communities"]["blackholing"]["std"] = c
            self._contains_err()
        for c in ("65536:666", "-1:-1", "0:-1", "65535:65536", "1", "1:1:1", "rt:1:0") + self.VALID_LRG_COMMS + self.VALID_EXT_COMMS:
            self.cfg["communities"]["blackholing"]["std"] = c
            self._contains_err("Invalid BGP standard community: {};".format(c))
        for c in ("65535:peer_as", "peer_as:1"):
            self.cfg["communities"]["blackholing"]["std"] = c
            self._contains_err("'peer_as' macro not allowed")
        self.cfg["communities"]["blackholing"]["std"] = self.VALID_STD_COMMS[0]
        self._test_optional(self.cfg["communities"]["blackholing"], "std")

    def test_communities_lrg(self):
        """Generic config parser: large BGP communities"""

        for c in self.VALID_LRG_COMMS:
            self.cfg["communities"]["blackholing"]["lrg"] = c
            self._contains_err()
        for c in ("4294967296:1:1", "-1:0:0") + self.VALID_STD_COMMS + self.VALID_EXT_COMMS:
            self.cfg["communities"]["blackholing"]["lrg"] = c
            self._contains_err("Invalid BGP large community: {};".format(c))
        for c in ("1:65535:peer_as", "peer_as:1:1", "1:peer_as:2"):
            self.cfg["communities"]["blackholing"]["lrg"] = c
            self._contains_err("'peer_as' macro not allowed")
        self.cfg["communities"]["blackholing"]["lrg"] = self.VALID_LRG_COMMS[0]
        self._test_optional(self.cfg["communities"]["blackholing"], "lrg")

    def test_communities_ext(self):
        """Generic config parser: extended BGP communities"""

        for c in self.VALID_EXT_COMMS:
            self.cfg["communities"]["blackholing"]["ext"] = c
            self._contains_err()
        for c in ("x:1:0", "ro:-1:0") + self.VALID_STD_COMMS + self.VALID_LRG_COMMS:
            self.cfg["communities"]["blackholing"]["ext"] = c
            self._contains_err("Invalid BGP extended community: {};".format(c))
        for c in ("rt:65535:peer_as", "ro:peer_as:1"):
            self.cfg["communities"]["blackholing"]["ext"] = c
            self._contains_err("'peer_as' macro not allowed")
        self._test_optional(self.cfg["communities"]["blackholing"], "ext")

    def test_fixed_values_communities(self):
        """Generic config parser: communities with fixed values"""

        for comm in ("announce_to_peer", "do_not_announce_to_peer"):
            for c in self.VALID_STD_COMMS:
                self.cfg["communities"][comm]["std"] = c
                self._contains_err("This community must have a fixed value:")
            for c in self.VALID_LRG_COMMS:
                self.cfg["communities"][comm]["lrg"] = c
                self._contains_err("This community must have a fixed value:")
            for c in self.VALID_EXT_COMMS:
                self.cfg["communities"][comm]["ext"] = c
                self._contains_err("This community must have a fixed value:")
            self.cfg["communities"][comm]["std"] = None
            self.cfg["communities"][comm]["lrg"] = None
            self.cfg["communities"][comm]["ext"] = None
            self.cfg.parse()

    def test_duplicate_communities(self):
        """Generic config parser: duplicate communities"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  communities:",
            "    prefix_present_in_as_set:",
            "      std: '999:1'",
            "    prefix_not_present_in_as_set:",
            "      lrg: 'rs_as:2:2'",
        ]
        yaml_lines = tpl + [
            "    origin_present_in_as_set:",
            "      std: 'rs_as:1'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("The 'prefix_present_in_as_set.std' community's value (999:1) has already been used for another community.")

        yaml_lines = tpl + [
            "    origin_present_in_as_set:",
            "      lrg: '999:2:2'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("The 'prefix_not_present_in_as_set.lrg' community's value (999:2:2) has already been used for another community.")

    def test_max_pref_action(self):
        """Generic config parser: max_prefix action"""
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["action"], None)
        self._test_option(self.cfg["filtering"]["max_prefix"], "action", ("shutdown", "restart", "block", "warning"))
        self._test_optional(self.cfg["filtering"]["max_prefix"], "action")

    def test_max_pref_peeringdb(self):
        """Generic config parser: max_prefix PeeringDB"""
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["peering_db"], True)
        self._test_bool_val(self.cfg["filtering"]["max_prefix"], "peering_db")
        self._test_optional(self.cfg["filtering"]["max_prefix"], "peering_db")

    def test_max_pref_general_limit_ipv4(self):
        """Generic config parser: max_prefix general_limit_ipv4"""
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["general_limit_ipv4"], 170000)
        self._test_optional(self.cfg["filtering"]["max_prefix"], "general_limit_ipv4")

    def test_max_pref_general_limit_ipv6(self):
        """Generic config parser: max_prefix general_limit_ipv6"""
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["general_limit_ipv6"], 12000)
        self._test_optional(self.cfg["filtering"]["max_prefix"], "general_limit_ipv6")

    def test_default_values(self):
        """Generic config parser: minimal config"""
        self.load_config(yaml="cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2"
        )
        self.cfg.parse()
        self._contains_err()

        self.assertEqual(self.cfg["prepend_rs_as"], False)
        self.assertEqual(self.cfg["path_hiding"], True)
        self.assertEqual(self.cfg["passive"], True)
        self.assertEqual(self.cfg["gtsm"], True)
        self.assertEqual(self.cfg["add_path"], False)
        self.assertEqual(self.cfg["filtering"]["next_hop_policy"], "strict")
        self.assertEqual(self.cfg["filtering"]["ipv4_pref_len"]["min"], 8)
        self.assertEqual(self.cfg["filtering"]["ipv4_pref_len"]["max"], 24)
        self.assertEqual(self.cfg["filtering"]["ipv6_pref_len"]["min"], 12)
        self.assertEqual(self.cfg["filtering"]["ipv6_pref_len"]["max"], 48)
        self.assertEqual(self.cfg["filtering"]["max_as_path_len"], 32)
        self.assertEqual(self.cfg["filtering"]["reject_invalid_as_in_as_path"], True)
        self.assertEqual(self.cfg["filtering"]["rpsl"]["tag_as_set"], True)
        self.assertEqual(self.cfg["filtering"]["rpsl"]["enforce_origin_in_as_set"], True)
        self.assertEqual(self.cfg["filtering"]["rpsl"]["enforce_prefix_in_as_set"], True)
        self.assertEqual(self.cfg["filtering"]["rpki"]["enabled"], False)
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["action"], None)
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["peering_db"], True)
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["general_limit_ipv4"], 170000)
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["general_limit_ipv6"], 12000)
        self.assertEqual(self.cfg["blackhole_filtering"]["policy_ipv4"], None)
        self.assertEqual(self.cfg["blackhole_filtering"]["policy_ipv6"], None)
        self.assertEqual(self.cfg["control_communities"], True)




# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ipaddr
from jinja2 import Environment, FileSystemLoader
import mock
import os
import time

from docker import InstanceError
from ..base import ARouteServerTestCase

from pierky.arouteserver.builder import BIRDConfigBuilder
from pierky.arouteserver.config.validators import ValidatorPrefixListEntry
from pierky.arouteserver.rpsl import ASSet, RSet


class LiveScenario(ARouteServerTestCase):

    MODULE_PATH = None
    SHORT_DESCR = ""

    IP_VER = None
    DATA = {}
    AS_SET = {}
    R_SET = {}
    INSTANCES = []

    DEBUG = False
    DO_NOT_STOP_INSTANCES = False

    CONFIG_BUILDER_CLASS = BIRDConfigBuilder

    @classmethod
    def _get_module_dir(cls):
        return os.path.dirname(cls.MODULE_PATH)

    @classmethod
    def _get_instance_by_name(cls, name):
        for instance in cls.INSTANCES:
            if instance.name == name:
                return instance
        return None

    @classmethod
    def _create_var_dir(cls):
        var_dir = "{}/var".format(cls._get_module_dir())
        try:
            os.mkdir(var_dir)
        except OSError:
            pass
        return var_dir

    @classmethod
    def _build_other_cfg(cls, tpl_name):
        cls.debug("Building config from {}/{} - IPv{}".format(
            cls._get_module_dir(), tpl_name, cls.IP_VER))

        env = Environment(
            loader=FileSystemLoader(cls._get_module_dir()),
            trim_blocks=True,
            lstrip_blocks=True
        )
        tpl = env.get_template(tpl_name)
        cfg = tpl.render(ip_ver=cls.IP_VER, data=cls.DATA)

        var_dir = cls._create_var_dir()
        cfg_file_path = "{}/{}.config".format(var_dir, tpl_name)

        with open(cfg_file_path, "w") as f:
            f.write(cfg)

        return cfg_file_path

    @classmethod
    def _build_rs_cfg(cls, tpl_dir_name, tpl_name, out_file_name,
                      cfg_general="general.yml", cfg_bogons="bogons.yml",
                      cfg_clients="clients.yml"):
        cls.debug("Building config from {}/{} - IPv{}".format(tpl_dir_name, tpl_name, cls.IP_VER))

        var_dir = cls._create_var_dir()

        builder = cls.CONFIG_BUILDER_CLASS(
            template_dir="{}/{}".format(cls._get_module_dir(), tpl_dir_name),
            template_name=tpl_name,
            cache_dir=var_dir,
            cfg_general="{}/{}".format(cls._get_module_dir(), cfg_general),
            cfg_bogons="{}/{}".format(cls._get_module_dir(), cfg_bogons),
            cfg_clients="{}/{}".format(cls._get_module_dir(), cfg_clients),
            ip_ver=cls.IP_VER
        )

        cfg_file_path = "{}/{}".format(var_dir, out_file_name)

        cfg = builder.render_template()
        with open(cfg_file_path, "w") as f:
            f.write(cfg)

        return cfg_file_path

    @classmethod
    def mock_rpsl(cls):
        def _mock_load_data_from_cache(*args, **kwargs):
            return False

        def _mock_RSet__get_data(self):
    
            def add_prefix_to_list(prefix_name, lst):
                obj = ipaddr.IPNetwork(cls.DATA[prefix_name])
                lst.append(
                    ValidatorPrefixListEntry().validate({
                        "prefix": str(obj.ip),
                        "length": obj.prefixlen,
                        "comment": self.object_name
                    })
                )

            res = []

            if self.object_name in cls.R_SET:
                for prefix_name in cls.R_SET[self.object_name]:
                    add_prefix_to_list(prefix_name, res)
            return res

        def _mock_ASSet__get_data(self):
            if self.object_name in cls.AS_SET:
                return cls.AS_SET[self.object_name]

        mock_RPSLTools_load_data_from_cache = mock.patch.object(
            RSet, "load_data_from_cache"
        ).start()
        mock_RPSLTools_load_data_from_cache.side_effect = _mock_load_data_from_cache

        mock_ASSet__get_data = mock.patch.object(
            ASSet, "_get_data", autospec=True
        ).start()
        mock_ASSet__get_data.side_effect = _mock_ASSet__get_data

        mock_RSet__get_data = mock.patch.object(
            RSet, "_get_data", autospec=True
        ).start()
        mock_RSet__get_data.side_effect = _mock_RSet__get_data

    @classmethod
    def _setUpClass(cls):
        print("Setting instances up...")

        cls.mock_rpsl()
        cls._setup_instances()
        try:
            for instance in cls.INSTANCES:
                instance.set_var_dir("{}/var".format(cls._get_module_dir()))

                if cls.DO_NOT_STOP_INSTANCES and instance.is_running():
                    cls.debug("Instance '{}' already running, reloading config".format(instance.name))
                    instance.remount()
                    if not instance.reload_config():
                        raise InstanceError("An error occurred while reloading '{}' configuration.".format(instance.name))
                    continue
                cls.debug("Starting instance '{}'...".format(instance.name))
                instance.start()
        except:
            cls.tearDownClass()
            raise

    @classmethod
    def _tearDownClass(cls):
        mock.patch.stopall()

        if cls.DO_NOT_STOP_INSTANCES:
            cls.debug("Skipping instances stopping")
            return
        print("Stopping instances...")
        for instance in cls.INSTANCES:
            cls.debug("Stopping instance '{}'...".format(instance.name))
            instance.stop()

    def set_instance_variables(self):
        raise NotImplementedError()

    def _setUp(self):
        self.set_instance_variables()

    def receive_route_from(self, inst, prefix, other_inst=None, as_path=None,
                           next_hop=None, std_comms=None, lrg_comms=None,
                           ext_comms=None, filtered=None):
        routes = inst.get_routes(prefix, include_filtered=filtered is not None)

        next_hop_ip = None
        if next_hop:
            next_hop_ip = next_hop if isinstance(next_hop, str) else next_hop.ip

        errors = []
        if not routes:
            errors.append("{inst} does not receive {prefix} at all.")
        else:
            for route in routes:
                err = False
                if other_inst and route.via != other_inst.ip:
                    errors.append("{{inst}} receives {{prefix}} from {via} and not from {{other_inst}}.".format(via=route.via))
                    err = True
                if as_path and route.as_path != as_path:
                    errors.append("{{inst}} receives {{prefix}} with AS_PATH {as_path} and not with {{as_path}}.".format(as_path=route.as_path))
                    err = True
                if next_hop and route.next_hop != next_hop_ip:
                    errors.append("{{inst}} receives {{prefix}} with NEXT_HOP {next_hop} and not with {{next_hop_ip}}.".format(next_hop=route.next_hop))
                    err = True
                if std_comms is not None and sorted(route.std_comms) != sorted(std_comms):
                    errors.append("{{inst}} receives {{prefix}} with std comms {comms} and not with {{std_comms}}.".format(comms=route.std_comms))
                    err = True
                if lrg_comms is not None and sorted(route.lrg_comms) != sorted(lrg_comms):
                    errors.append("{{inst}} receives {{prefix}} with lrg comms {comms} and not with {{lrg_comms}}.".format(comms=route.lrg_comms))
                    err = True
                if ext_comms is not None and sorted(route.ext_comms) != sorted(ext_comms):
                    errors.append("{{inst}} receives {{prefix}} with ext comms {comms} and not with {{ext_comms}}.".format(comms=route.ext_comms))
                    err = True
                if filtered is not None and route.filtered != filtered:
                    errors.append(
                        "{{inst}} receives {{prefix}} from {via}, AS_PATH {as_path}, NEXT_HOP {next_hop} "
                        "but it is {filtered_status} while it is expected to be {filtered_exp}.".format(
                            via=route.via,
                            as_path=route.as_path,
                            next_hop=route.next_hop,
                            filtered_status="filtered" if route.filtered else "not filtered",
                            filtered_exp="filtered" if filtered else "not filtered"
                        )
                    )
                    err = True
                if not err:
                    return

        if errors:
            criteria = []
            if other_inst:
                criteria.append("from {} ({})".format(other_inst.ip, other_inst.name))
            if as_path:
                criteria.append("with AS_PATH {}".format(as_path))
            if next_hop_ip:
                criteria.append("with next-hop {}".format(next_hop_ip))
            if std_comms:
                criteria.append("with std comms {}".format(std_comms))
            if lrg_comms:
                criteria.append("with lrg comms {}".format(lrg_comms))
            if ext_comms:
                criteria.append("with ext comms {}".format(ext_comms))
            if filtered is True:
                criteria.append("filtered")

            failure = "Looking for {prefix} on {inst} {criteria}:\n\t".format(
                prefix=prefix,
                inst=inst.name,
                criteria=", ".join(criteria)
            )
            failure += "\n\t".join([
                err.format(
                    inst=inst.name,
                    prefix=prefix,
                    other_inst="{} ({})".format(other_inst.ip, other_inst.name) if other_inst else "",
                    as_path=as_path,
                    next_hop_ip=next_hop_ip,
                    std_comms=std_comms,
                    lrg_comms=lrg_comms,
                    ext_comms=ext_comms,
                ) for err in errors
            ])
            self.fail(failure)

    def log_contains(self, inst, msg, instances={}):
        expanded_msg = msg

        macros_dict = {}
        for inst_name in instances:
            other_inst = instances[inst_name]
            # BAD THING: using a BIRD specific method to understand
            # what's the other instance's name on 'inst'.
            proto_name = inst.get_protocol_name_by_ip(other_inst.ip)
            macros_dict[inst_name] = proto_name

        if macros_dict:
            expanded_msg = expanded_msg.format(**macros_dict)

        if not inst.log_contains(expanded_msg):
            self.fail("Expected message not found on {} logs:\n\t{}".format(inst.name, expanded_msg))

    def session_is_up(self, inst_a, inst_b):
        if inst_a.bgp_session_is_up(inst_b):
            return
        time.sleep(5)
        if inst_a.bgp_session_is_up(inst_b, force_update=True):
            return
        self.fail("BGP session between '{}' ({}) and '{}' ({}) is not up.".format(
            inst_a.name, inst_a.ip, inst_b.name, inst_b.ip
        ))



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .base import DefaultConfigScenario
from ...bird import BIRDInstanceIPv6

class TagASSetScenario_BIRDIPv6(DefaultConfigScenario):
    __test__ = True

    SHORT_DESCR = "Live test, BIRD, examples, rich config, IPv6"
    RS_INSTANCE_CLASS = BIRDInstanceIPv6
    CLIENT_INSTANCE_CLASS = BIRDInstanceIPv6
    IP_VER = 6

    DATA = {
        "rs_IPAddress":                     "2001:db8:1:1::2",

        "AS10745_allowed_prefixes":         "2001:500:4::/48",
        "AS3333_allowed_prefixes":          "2001:67c:2e8::/48",
    }



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from docker import DockerInstance
from instances import Route


class BIRDInstance(DockerInstance):

    DOCKER_IMAGE = "pierky/bird:1.6.3"

    def __init__(self, *args, **kwargs):
        DockerInstance.__init__(self, *args, **kwargs)
        self.protocols_status = {}

    def reload_config(self):
        if not self.is_running():
            raise InstanceNotRunning(self.name)

        res = self._birdcl("configure")

        if "Reconfigured" in res:
            return True
        else:
            return False

    def _get_protocols_status(self, force_update=False):
        if force_update:
            self.protocols_status = {}
        if self.protocols_status:
            return
        out = self._birdcl("show protocols all")
        lines = out.split("\n")
        last_protoname = None
        last_proto_status = None
        for line in lines:
            if line.strip() and not line.startswith(" "):
                last_protoname = line.split(" ")[0]
                last_proto_status = re.search("\sup\s", line) is not None
                continue
            if line.strip().startswith("Neighbor address: "):
                neighbor_ip = line.strip().split(": ")[1]
                if last_protoname is not None and last_proto_status is not None:
                    self.protocols_status[last_protoname] = {
                        "ip": neighbor_ip,
                        "is_up": last_proto_status
                    }
                    last_protoname = None
                    last_proto_status = None
                else:
                    raise Exception(
                        "{}: neighbor with unknown protocol name/status: {}".format(
                            self.name, neighbor_ip
                        )
                    )
        if not self.protocols_status:
            raise Exception("{}: can't build protocols status map".format(self.name))

    def bgp_session_is_up(self, other_inst, force_update=False):
        self._get_protocols_status(force_update=force_update)
        for proto in self.protocols_status:
            if self.protocols_status[proto]["ip"] == other_inst.ip:
                return self.protocols_status[proto]["is_up"]
        raise Exception("Can't get BGP session status for {}".format(other_inst.name))

    def get_routes(self, prefix, include_filtered=False):
        self._get_protocols_status()
        routes = []

        regex = "[ ]+via ([0-9\.\:a-f]+)[^\[]+\[([^\s]+)[^\n]+"

        for filtered in ["", "filtered"] if include_filtered else [""]:
            cmd = "show route {} all {}".format(prefix, filtered)
            out = self._birdcl(cmd)
            match = re.search("^[0-9\.\:a-f]+/[0-9]+{}".format(regex), out, re.MULTILINE)
            if match:
                route = {}
                lines = out.split("\n")
                for line in lines:
                    if not line.strip():
                        continue
                    match = re.search(regex, line)
                    if match:
                        if route:
                            route["filtered"] = filtered == "filtered"
                            routes.append(Route(**route))
                            route = {}
                        route["prefix"] = prefix
                        route["via"] = self.protocols_status[match.group(2)]["ip"]
                    else:
                        if "BGP.as_path:" in line:
                            route["as_path"] = line.split(": ")[1].strip()
                        if "BGP.next_hop:" in line:
                            route["next_hop"] = line.split(": ")[1].strip()
                        if "BGP.community:" in line:
                            route["std_comms"] = line.split(": ")[1].strip()
                        if "BGP.large_community:" in line:
                            route["lrg_comms"] = line.split(": ")[1].strip()
                route["filtered"] = filtered == "filtered"
                routes.append(Route(**route))
        return routes

    def get_protocol_name_by_ip(self, ip):
        self._get_protocols_status()
        for proto in self.protocols_status:
            if self.protocols_status[proto]["ip"] == ip:
                return proto
        raise Exception("{}: can't find protocolo name from ip {}".format(self.name, ip))
        return None

    def log_contains(self, s):
        out = self.run_cmd("cat /var/log/bird.log")
        if s in out:
            return True
        else:
            return False

class BIRDInstanceIPv4(BIRDInstance):

    def _get_start_cmd(self):
        return "bird -c /etc/bird/bird.conf -d"

    def _birdcl(self, cmd):
        return self.run_cmd("birdcl {}".format(cmd))

class BIRDInstanceIPv6(BIRDInstance):

    def _get_start_cmd(self):
        return "bird6 -c /etc/bird/bird.conf -d"

    def _birdcl(self, cmd):
        return self.run_cmd("birdcl6 {}".format(cmd))




# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ...base import LiveScenario

class BasicScenario(LiveScenario):
    __test__ = False

    MODULE_PATH = __file__
    RS_INSTANCE_CLASS = None
    CLIENT_INSTANCE_CLASS = None
    IP_VER = None

    AS_SET = {
        "AS-AS1": [1],
        "AS-AS1_CUSTOMERS": [101],
        "AS-AS2": [2],
        "AS-AS2_CUSTOMERS": [101],
    }
    R_SET = {
        "AS-AS1": [
            "AS1_allowed_prefixes",
            "pref_len1"
        ],
        "AS-AS1_CUSTOMERS": [
            "AS101_allowed_prefixes"
        ],
        "AS-AS2": [
            "AS2_allowed_prefixes"
        ],
        "AS-AS2_CUSTOMERS": [
            "AS101_allowed_prefixes"
        ],
    }

    @classmethod
    def _setup_instances(cls):
        cls.INSTANCES = [
            cls.RS_INSTANCE_CLASS(
                "rs",
                cls.DATA["rs_IPAddress"],
                [
                    (
                        cls._build_rs_cfg("bird", "main.j2", "rs.conf"),
                        "/etc/bird/bird.conf"
                    )
                ],
                proto_name="the_rs"
            ),
            # Run AS3 as soon as possible because it's configured with passive
            # session, so while other instance come up rs has more time to
            # setup the session with it.
            cls.CLIENT_INSTANCE_CLASS(
                "AS3",
                cls.DATA["AS3_1_IPAddress"],
                [
                    (
                        cls._build_other_cfg("AS3.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
                client_id="AS3_1"
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS1_1",
                cls.DATA["AS1_1_IPAddress"],
                [
                    (
                        cls._build_other_cfg("AS1_1.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
                client_id="AS1_1"
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS1_2",
                cls.DATA["AS1_2_IPAddress"],
                [
                    (
                        cls._build_other_cfg("AS1_2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
                client_id="AS1_2"
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS2",
                cls.DATA["AS2_1_IPAddress"],
                [
                    (
                        cls._build_other_cfg("AS2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
                client_id="AS2_1"
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS101",
                cls.DATA["AS101_IPAddress"],
                [
                    (
                        cls._build_other_cfg("AS101.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
                proto_name="AS101"
            )
        ]

    def set_instance_variables(self):
        self.AS1_1 = self._get_instance_by_name("AS1_1")
        self.AS1_2 = self._get_instance_by_name("AS1_2")
        self.AS2 = self._get_instance_by_name("AS2")
        # AS3 is passive, rs client configured with passive: False
        self.AS3 = self._get_instance_by_name("AS3")
        self.AS101 = self._get_instance_by_name("AS101")
        self.rs = self._get_instance_by_name("rs")

    def test_010_setup(self):
        """{}: instances setup"""
        pass

    def test_020_sessions_up(self):
        """{}: sessions are up"""
        self.session_is_up(self.rs, self.AS1_1)
        self.session_is_up(self.rs, self.AS1_2)
        self.session_is_up(self.rs, self.AS2)
        self.session_is_up(self.rs, self.AS3)
        self.session_is_up(self.AS101, self.AS1_1)
        self.session_is_up(self.AS101, self.AS1_2)
        self.session_is_up(self.AS101, self.AS2)

    def test_030_good_prefixes_received_by_rs(self):
        """{}: good prefixes received by rs"""
        self.receive_route_from(self.rs, self.DATA["AS1_good1"], self.AS1_1)
        self.receive_route_from(self.rs, self.DATA["AS1_good2"], self.AS1_1)
        self.receive_route_from(self.rs, self.DATA["AS1_good1"], self.AS1_2)
        self.receive_route_from(self.rs, self.DATA["AS1_good2"], self.AS1_2)
        # AS1_good3 is announced by AS1_2 with NEXT_HOP = AS1_1
        self.receive_route_from(self.rs, self.DATA["AS1_good3"], self.AS1_2, next_hop=self.AS1_1)
        self.receive_route_from(self.rs, self.DATA["AS2_good1"], self.AS2)
        self.receive_route_from(self.rs, self.DATA["AS2_good2"], self.AS2)

        # rs should not receive prefixes with the following criteria
        with self.assertRaises(AssertionError):
            self.receive_route_from(self.rs, self.DATA["AS1_good1"], self.AS2)
        with self.assertRaises(AssertionError):
            self.receive_route_from(self.rs, self.DATA["AS1_good2"], self.AS2)
        with self.assertRaises(AssertionError):
            self.receive_route_from(self.rs, self.DATA["AS2_good1"], self.AS1_1)
        with self.assertRaises(AssertionError):
            self.receive_route_from(self.rs, self.DATA["AS2_good2"], self.AS1_1)

        # AS_PATH should match the expectations
        self.receive_route_from(self.rs, self.DATA["AS1_good1"], self.AS1_1, as_path="1")
        self.receive_route_from(self.rs, self.DATA["AS2_good1"], self.AS2, as_path="2")

        # NEXT_HOP should match the expectations
        self.receive_route_from(self.rs, self.DATA["AS1_good1"], self.AS1_1, as_path="1", next_hop=self.AS1_1)
        self.receive_route_from(self.rs, self.DATA["AS2_good1"], self.AS2, as_path="2", next_hop=self.AS2)

    def test_040_bad_prefixes_received_by_rs_bogon(self):
        """{}: bad prefixes received by rs: bogon"""
        self.receive_route_from(self.rs, self.DATA["bogon1"], self.AS1_1, as_path="1", next_hop=self.AS1_1, filtered=True)
        self.log_contains(self.rs, "prefix is bogon - REJECTING " + self.DATA["bogon1"])

    def test_040_bad_prefixes_received_by_rs_global_blacklist(self):
        """{}: bad prefixes received by rs: global blacklist"""

        self.receive_route_from(self.rs, self.DATA["local1"], self.AS1_1, as_path="1", next_hop=self.AS1_1, filtered=True)
        self.log_contains(self.rs, "prefix is in global blacklist - REJECTING " + self.DATA["local1"])

    def test_040_bad_prefixes_received_by_rs_prefix_len(self):
        """{}: bad prefixes received by rs: invalid prefix-len"""

        self.receive_route_from(self.rs, self.DATA["pref_len1"], self.AS1_1, as_path="1", next_hop=self.AS1_1, filtered=True)
        pref_len = 7 if self.IP_VER == 4 else 16
        pref_len_range = "8-24" if self.IP_VER == 4 else "17-48"
        self.log_contains(self.rs, "prefix len [{}] not in {} - REJECTING {}".format(
            pref_len, pref_len_range, self.DATA["pref_len1"]))

    def test_040_bad_prefixes_received_by_rs_leftmost_asn(self):
        """{}: bad prefixes received by rs: left-most ASN"""

        self.receive_route_from(self.rs, self.DATA["peer_as1"], self.AS1_1, as_path="2 1", next_hop=self.AS1_1, filtered=True)
        self.log_contains(self.rs, "invalid left-most ASN [2] - REJECTING " + self.DATA["peer_as1"])

    def test_040_bad_prefixes_received_by_rs_invalid_aspath(self):
        """{}: bad prefixes received by rs: invalid ASN in AS-PATH"""

        self.receive_route_from(self.rs, self.DATA["invalid_asn1"], self.AS1_1, as_path="1 65536 1", next_hop=self.AS1_1, filtered=True)
        self.log_contains(self.rs, "AS_PATH [(path 1 65536 1)] contains invalid ASN - REJECTING " + self.DATA["invalid_asn1"])

    def test_040_bad_prefixes_received_by_rs_aspath_len(self):
        """{}: bad prefixes received by rs: AS_PATH len"""

        self.receive_route_from(self.rs, self.DATA["aspath_len1"], self.AS1_1, as_path="1 2 2 2 2 2 2 1", next_hop=self.AS1_1, filtered=True)
        self.log_contains(self.rs, "AS_PATH len [8] longer than 6 - REJECTING " + self.DATA["aspath_len1"])

    def test_040_bad_prefixes_received_by_rs_client_blacklist(self):
        """{}: bad prefixes received by rs: client blacklist"""

        self.receive_route_from(self.rs, self.DATA["AS3_blacklist1"], self.AS3, as_path="3", next_hop=self.AS3, filtered=True)
        self.log_contains(self.rs, "prefix is in client's blacklist - REJECTING " + self.DATA["AS3_blacklist1"])

    def test_040_bad_prefixes_received_by_rs_invalid_nexthop(self):
        """{}: bad prefixes received by rs: invalid NEXT_HOP"""

        self.receive_route_from(self.rs, self.DATA["AS101_good1"], self.AS1_2, as_path="1 101", next_hop=self.AS101, filtered=True)
        self.log_contains(self.rs, "NEXT_HOP [" + self.AS101.ip + "] not allowed - REJECTING " + self.DATA["AS101_good1"])

    def test_040_bad_prefixes_received_by_rs_no_rset(self):
        """{}: bad prefixes received by rs: prefix not in AS-SET"""

        # AS101_no_rset is not included in AS-AS1_CUSTOMERS nor in AS-AS2_CUSTOMERS, so it's
        # rejected by the rs.
        self.receive_route_from(self.rs, self.DATA["AS101_no_rset"], self.AS1_1, as_path="1 101", next_hop=self.AS1_1, filtered=True)
        self.receive_route_from(self.rs, self.DATA["AS101_no_rset"], self.AS2, as_path="2 101", next_hop=self.AS2, filtered=True)
        self.log_contains(self.rs, "prefix not in client's r_set - REJECTING " + self.DATA["AS101_no_rset"])

    def test_040_bad_prefixes_received_by_rs_no_asset(self):
        """{}: bad prefixes received by rs: origin not in AS-SET"""

        # AS102_no_asset is announced by an (hypothetical) AS102 to AS101,
        # and AS102 is not included in AS-AS1_CUSTOMERS nor in AS-AS2_CUSTOMERS, so it's
        # rejected by the rs.
        self.receive_route_from(self.rs, self.DATA["AS102_no_asset"], self.AS1_1, as_path="1 101 102", next_hop=self.AS1_1, filtered=True)
        self.receive_route_from(self.rs, self.DATA["AS102_no_asset"], self.AS2, as_path="2 101 102", next_hop=self.AS2, filtered=True)
        self.log_contains(self.rs, "origin ASN [102] not in allowed as-sets - REJECTING " + self.DATA["AS102_no_asset"])

    def test_041_bad_prefixes_not_received_by_clients(self):
        """{}: bad prefixes not received by clients"""
        for prefix in (self.DATA["bogon1"],
                       self.DATA["local1"],
                       self.DATA["pref_len1"],
                       self.DATA["peer_as1"],
                       self.DATA["invalid_asn1"],
                       self.DATA["aspath_len1"]):
            for inst in (self.AS2, self.AS3):
                with self.assertRaises(AssertionError):
                    self.receive_route_from(inst, prefix)

        # Among the clients, only AS3 is expected to not see the 
        # following prefixes because AS1 and AS2
        # receive them on their session with AS101
        for prefix in (self.DATA["AS101_no_rset"],
                       self.DATA["AS102_no_asset"]):
            with self.assertRaises(AssertionError):
                self.receive_route_from(self.AS3, prefix)

    def test_050_prefixes_from_AS101_received_by_its_upstreams(self):
        """{}: prefixes from AS101 received by its upstreams"""
        self.receive_route_from(self.AS1_1, self.DATA["AS101_good1"], self.AS101)
        self.receive_route_from(self.AS1_2, self.DATA["AS101_good1"], self.AS101)
        self.receive_route_from(self.AS2, self.DATA["AS101_good1"], self.AS101)

    def test_051_prefixes_from_AS101_received_by_rs(self):
        """{}: prefixes from AS101 received by rs"""

        # rs should receive these prefixes that AS101 announces to AS1 and AS2.
        self.receive_route_from(self.rs, self.DATA["AS101_good1"], self.AS1_1, as_path="1 101", next_hop=self.AS1_1)
        self.receive_route_from(self.rs, self.DATA["AS101_good1"], self.AS2, as_path="2 101", next_hop=self.AS2)

        # AS101 peers with AS1_2 on the same network of rs
        # and AS1_2 has not 'next-hop-self' on the session with rs:
        # next-hop received by rs is the one of AS101, so the prefix
        # should be filtered by the rs.
        self.receive_route_from(self.rs, self.DATA["AS101_good1"], self.AS1_2, as_path="1 101", next_hop=self.AS101, filtered=True)

    def test_060_communities_as_seen_by_AS101_upstreams(self):
        """{}: bad communities as seen by AS101 upstreams"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            self.receive_route_from(inst, self.DATA["AS101_bad_std_comm"], self.AS101, std_comms=["65530:0"])
            self.receive_route_from(inst, self.DATA["AS101_bad_lrg_comm"], self.AS101, lrg_comms=["999:65530:0"])
            self.receive_route_from(inst, self.DATA["AS101_other_s_comm"], self.AS101, std_comms=["888:0"])
            self.receive_route_from(inst, self.DATA["AS101_other_l_comm"], self.AS101, lrg_comms=["888:0:0"])
            self.receive_route_from(inst, self.DATA["AS101_bad_good_comms"], self.AS101, std_comms=["777:0", "65530:1"], lrg_comms=["777:0:0", "999:65530:1"])

    def test_061_bad_communities_scrubbed_by_rs(self):
        """{}: bad communities scrubbed by rs"""
        
        self.receive_route_from(self.rs, self.DATA["AS101_bad_std_comm"], self.AS1_1, std_comms=[])
        self.receive_route_from(self.rs, self.DATA["AS101_bad_std_comm"], self.AS2, std_comms=[])
        self.receive_route_from(self.rs, self.DATA["AS101_bad_lrg_comm"], self.AS1_1, std_comms=[])
        self.receive_route_from(self.rs, self.DATA["AS101_bad_lrg_comm"], self.AS2, std_comms=[])

    def test_062_other_communities_not_scrubbed_by_rs(self):
        """{}: other communities not scrubbed by rs"""

        for inst in (self.AS1_1, self.AS2):
            self.receive_route_from(self.rs, self.DATA["AS101_other_s_comm"], inst, std_comms=["888:0"])
            self.receive_route_from(self.rs, self.DATA["AS101_other_l_comm"], inst, lrg_comms=["888:0:0"])
            self.receive_route_from(self.rs, self.DATA["AS101_bad_good_comms"], inst, std_comms=["777:0"], lrg_comms=["777:0:0"])

    def test_070_blackhole_filtering_as_seen_by_rs(self):
      """{}: blackhole filtering requests as seen by rs"""

      self.receive_route_from(self.rs, self.DATA["AS2_blackhole1"], self.AS2, next_hop=self.AS2, as_path="2",
                              std_comms=["65535:666"], lrg_comms=[])
      self.log_contains(self.rs, "blackhole filtering request from {AS2_1} - ACCEPTING " + self.DATA["AS2_blackhole1"], {"AS2_1": self.AS2})

      self.receive_route_from(self.rs, self.DATA["AS2_blackhole2"], self.AS2, next_hop=self.AS2, as_path="2",
                              std_comms=["65534:0"], lrg_comms=[])
      self.log_contains(self.rs, "blackhole filtering request from {AS2_1} - ACCEPTING " + self.DATA["AS2_blackhole2"], {"AS2_1": self.AS2})

      self.receive_route_from(self.rs, self.DATA["AS2_blackhole3"], self.AS2, next_hop=self.AS2, as_path="2",
                              std_comms=[], lrg_comms=["65534:0:0"])
      self.log_contains(self.rs, "blackhole filtering request from {AS2_1} - ACCEPTING " + self.DATA["AS2_blackhole3"], {"AS2_1": self.AS2})

    def test_071_blackholed_prefixes_as_seen_by_enabled_clients(self):
      """{}: blackholed prefixes as seen by enabled clients"""

      for inst in (self.AS1_1, self.AS3):
        self.receive_route_from(inst, self.DATA["AS2_blackhole1"], self.rs, next_hop=self.DATA["blackhole_IP"],
            std_comms=["65535:666"], lrg_comms=[])
        self.receive_route_from(inst, self.DATA["AS2_blackhole2"], self.rs, next_hop=self.DATA["blackhole_IP"],
            std_comms=["65535:666"], lrg_comms=[])
        self.receive_route_from(inst, self.DATA["AS2_blackhole3"], self.rs, next_hop=self.DATA["blackhole_IP"],
            std_comms=["65535:666"], lrg_comms=[])

    def test_71_blackholed_prefixes_not_seen_by_not_enabled_clients(self):
        """{}: blackholed prefixes not seen by not enabled clients"""

        # AS1_2 not enabled to receive blackholed prefixes
        with self.assertRaises(AssertionError):
            self.receive_route_from(self.AS1_2, self.DATA["AS2_blackhole1"])
            self.receive_route_from(self.AS1_2, self.DATA["AS2_blackhole2"])
            self.receive_route_from(self.AS1_2, self.DATA["AS2_blackhole3"])
        proto_name = self.rs.get_protocol_name_by_ip(self.AS1_2.ip)
        self.log_contains(self.rs, "client {{AS1_2}} not enabled to receive blackhole prefixes - NOT ANNOUNCING {pref} TO {{AS1_2}}".format(pref=self.DATA["AS2_blackhole1"]), {"AS1_2": self.AS1_2})
        self.log_contains(self.rs, "client {{AS1_2}} not enabled to receive blackhole prefixes - NOT ANNOUNCING {pref} TO {{AS1_2}}".format(pref=self.DATA["AS2_blackhole2"]), {"AS1_2": self.AS1_2})
        self.log_contains(self.rs, "client {{AS1_2}} not enabled to receive blackhole prefixes - NOT ANNOUNCING {pref} TO {{AS1_2}}".format(pref=self.DATA["AS2_blackhole3"]), {"AS1_2": self.AS1_2})

    def test_080_control_communities_AS1_only(self):
        """{}: control communities, announce to AS1 only"""

        self.receive_route_from(self.AS1_1, self.DATA["AS3_cc_AS1only"], self.rs,
                                as_path="3", next_hop=self.AS3,
                                std_comms=[], lrg_comms=[])
        self.receive_route_from(self.AS1_2, self.DATA["AS3_cc_AS1only"], self.rs,
                                as_path="3", next_hop=self.AS3,
                                std_comms=[], lrg_comms=[])

        with self.assertRaises(AssertionError):
            self.receive_route_from(self.AS2, self.DATA["AS3_cc_AS1only"])
        self.log_contains(self.rs, "prefix didn't pass control communities checks - NOT ANNOUNCING {} TO {{AS2}}".format(self.DATA["AS3_cc_AS1only"]), {"AS2": self.AS2})

    def test_080_control_communities_not_AS1(self):
        """{}: control communities, announce to all except AS1"""

        self.receive_route_from(self.AS2, self.DATA["AS3_cc_not_AS1"], self.rs,
                                as_path="3", next_hop=self.AS3,
                                std_comms=[], lrg_comms=[])

        for inst in (self.AS1_1, self.AS1_2):
            with self.assertRaises(AssertionError):
                self.receive_route_from(inst, self.DATA["AS3_cc_not_AS1"])
            self.log_contains(self.rs, "prefix didn't pass control communities checks - NOT ANNOUNCING {} TO {{other_inst}}".format(self.DATA["AS3_cc_not_AS1"]), {"other_inst": inst})

    def test_080_control_communities_none(self):
        """{}: control communities, don't announce to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            with self.assertRaises(AssertionError):
                self.receive_route_from(inst, self.DATA["AS3_cc_none"])
            self.log_contains(self.rs, "prefix didn't pass control communities checks - NOT ANNOUNCING {} TO {{other_inst}}".format(self.DATA["AS3_cc_none"]), {"other_inst": inst})

    def test_081_control_communities_prepend1any(self):
        """{}: control communities, prepend once to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            self.receive_route_from(inst, self.DATA["AS3_prepend1any"], self.rs, as_path="3 3",
                                    next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_081_control_communities_prepend2any(self):
        """{}: control communities, prepend twice to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            self.receive_route_from(inst, self.DATA["AS3_prepend2any"], self.rs, as_path="3 3 3",
                                    next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_081_control_communities_prepend3any(self):
        """{}: control communities, prepend thrice to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            self.receive_route_from(inst, self.DATA["AS3_prepend3any"], self.rs, as_path="3 3 3 3",
                                    next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_100_prefixes_received_by_clients_AS1_1(self):
        """{}: prefixes received by clients: AS1_1"""
        self.receive_route_from(self.AS1_1, self.DATA["AS2_good1"], self.rs, as_path="2", next_hop=self.AS2)
        self.receive_route_from(self.AS1_1, self.DATA["AS2_good2"], self.rs, as_path="2", next_hop=self.AS2)

    def test_100_prefixes_received_by_clients_AS1_2(self):
        """{}: prefixes received by clients: AS1_2"""

        self.receive_route_from(self.AS1_2, self.DATA["AS2_good1"], self.rs, as_path="2", next_hop=self.AS2)
        self.receive_route_from(self.AS1_2, self.DATA["AS2_good2"], self.rs, as_path="2", next_hop=self.AS2)

    def test_100_prefixes_received_by_clients_AS2(self):
        """{}: prefixes received by clients: AS2"""

        self.receive_route_from(self.AS2, self.DATA["AS1_good1"], self.rs, as_path="1", next_hop=self.AS1_1)
        self.receive_route_from(self.AS2, self.DATA["AS1_good2"], self.rs, as_path="1", next_hop=self.AS1_1)
        self.receive_route_from(self.AS2, self.DATA["AS1_good3"], self.rs, as_path="1", next_hop=self.AS1_1)

    def test_100_prefixes_received_by_clients_AS3(self):
        """{}: prefixes received by clients: AS3 (with ADD-PATH)"""
        
        # AS3 has prepend_rs_as, so prefixes have AS_PATH "999 x"

        # AS3 must receive prefixes from both AS1_1 and AS1_2
        self.receive_route_from(self.AS3, self.DATA["AS1_good1"], self.rs, as_path="999 1", next_hop=self.AS1_1)
        self.receive_route_from(self.AS3, self.DATA["AS1_good1"], self.rs, as_path="999 1", next_hop=self.AS1_2)
        self.receive_route_from(self.AS3, self.DATA["AS1_good2"], self.rs, as_path="999 1", next_hop=self.AS1_1)
        self.receive_route_from(self.AS3, self.DATA["AS1_good2"], self.rs, as_path="999 1", next_hop=self.AS1_2)

        # prefix announced by AS1_2 only but with AS1_1 next-hop
        self.receive_route_from(self.AS3, self.DATA["AS1_good3"], self.rs, as_path="999 1", next_hop=self.AS1_1)

        # AS3 must receive prefixes from AS2
        self.receive_route_from(self.AS3, self.DATA["AS2_good1"], self.rs, as_path="999 2", next_hop=self.AS2)
        self.receive_route_from(self.AS3, self.DATA["AS2_good2"], self.rs, as_path="999 2", next_hop=self.AS2)

        # AS101 announces its prefixes to both AS1 (AS1_1 and AS1_2 clients) and AS2.
        # AS101 prefixes received by AS1_2 are rejected by rs because of next-hop-policy failure.
        # The other prefixes should be received twice by AS3.
        self.receive_route_from(self.AS3, self.DATA["AS101_good1"], self.rs, as_path="999 1 101", next_hop=self.AS1_1)
        self.receive_route_from(self.AS3, self.DATA["AS101_good1"], self.rs, as_path="999 2 101", next_hop=self.AS2)

        self.receive_route_from(self.AS3, self.DATA["AS101_bad_std_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1)
        self.receive_route_from(self.AS3, self.DATA["AS101_bad_std_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2)

        self.receive_route_from(self.AS3, self.DATA["AS101_bad_lrg_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1)
        self.receive_route_from(self.AS3, self.DATA["AS101_bad_lrg_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2)

        self.receive_route_from(self.AS3, self.DATA["AS101_other_s_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1, std_comms=["888:0"])
        self.receive_route_from(self.AS3, self.DATA["AS101_other_s_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2, std_comms=["888:0"])

        self.receive_route_from(self.AS3, self.DATA["AS101_other_l_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1, lrg_comms=["888:0:0"])
        self.receive_route_from(self.AS3, self.DATA["AS101_other_l_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2, lrg_comms=["888:0:0"])

        self.receive_route_from(self.AS3, self.DATA["AS101_bad_good_comms"], self.rs, as_path="999 1 101", next_hop=self.AS1_1,
                                std_comms=["777:0"], lrg_comms=["777:0:0"])
        self.receive_route_from(self.AS3, self.DATA["AS101_bad_good_comms"], self.rs, as_path="999 2 101", next_hop=self.AS2,
                                std_comms=["777:0"], lrg_comms=["777:0:0"])





# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ipaddr
import logging
import os

from jinja2 import Environment, FileSystemLoader

from .errors import MissingDirError, MissingFileError, BuilderError, \
                    ARouteServerError, PeeringDBError, PeeringDBNoInfoError, \
                    MissingArgumentError
from .config.general import ConfigParserGeneral
from .config.bogons import ConfigParserBogons
from .config.clients import ConfigParserClients
from .rpsl import ASSet, RSet
from .peering_db import PeeringDBNet


class ConfigBuilder(object):

    def validate_flavor_specific_configuration(self):
        raise NotImplementedError()

    @staticmethod
    def _get_cfg(obj_or_path, cls, descr, **kwargs):
        assert obj_or_path is not None
        if isinstance(obj_or_path, cls):
            return obj_or_path
        elif isinstance(obj_or_path, str):
            if not os.path.isfile(obj_or_path):
                raise MissingFileError(obj_or_path)
            obj = cls(**kwargs)
            try:
                obj.load(obj_or_path)
            except ARouteServerError as e:
                raise BuilderError(
                    "One or more errors occurred while loading "
                    "{} configuration file{}".format(
                        descr, ": {}".format(str(e)) if str(e) else ""
                    )
                )
            return obj

    @staticmethod
    def _check_is_dir(arg, path):
        if path is None:
            raise MissingArgumentError(arg)
        if not os.path.isdir(path):
            raise MissingDirError(path)
        return path

    def __init__(self, **kwargs):

        self.template_dir = self._check_is_dir(
            "template_dir", kwargs.get("template_dir")
        )

        self.template_name = kwargs.get("template_name")
        if not self.template_name:
            raise MissingArgumentError("template_name")

        self.template_path = os.path.join(self.template_dir,
                                          self.template_name)
        if not os.path.isfile(self.template_path):
            raise MissingFileError(self.template_path)

        self.cache_dir = self._check_is_dir(
            "cache_dir", kwargs.get("cache_dir")
        )

        try:
            with open(os.path.join(self.cache_dir, "write_test"), "w") as f:
                f.write("OK")
        except Exception as e:
            raise BuilderError(
                "Can't write into cache dir {}: {}".format(
                    self.cache_dir, str(e)
                )
            )

        self.ip_ver = kwargs.get("ip_ver", None)
        if self.ip_ver is not None:
            self.ip_ver = int(self.ip_ver)
            if self.ip_ver not in (4, 6):
                raise BuilderError("Invalid IP version: {}".format(ip_ver))

        self.cfg_general = self._get_cfg(kwargs.get("cfg_general"),
                                         ConfigParserGeneral,
                                         "general")
        self.cfg_bogons = self._get_cfg(kwargs.get("cfg_bogons"),
                                        ConfigParserBogons,
                                        "bogons")
        self.cfg_clients = self._get_cfg(kwargs.get("cfg_clients"),
                                         ConfigParserClients,
                                         "clients",
                                         general_cfg=self.cfg_general)

        self.validate_flavor_specific_configuration()

        self.enrich_config()

    def enrich_config_rpsl_as_set(self, client_id, as_sets, dest_list):
        errors = False
        for as_set in as_sets:
            try:
                asns = ASSet(as_set, cache_dir=self.cache_dir).asns
                if not asns:
                    raise BuilderError("it's empty")
                dest_list.extend(
                    [asn for asn in asns if asn not in dest_list]
                )
            except ARouteServerError as e:
                errors = True
                logging.error(
                    "Error while retrieving as_set {} for {}: {}".format(
                        as_set, client_id, str(e)
                    )
                )

        if errors:
            raise BuilderError()

    def enrich_config_rpsl_r_set(self, client_id, as_sets, dest_list):
        errors = False
        ip_versions = [self.ip_ver] if self.ip_ver else [4, 6]
        for as_set in as_sets:
            for ip_ver in ip_versions:
                try:
                    prefixes = RSet(as_set, ip_ver,
                                    cache_dir=self.cache_dir).prefixes
                    if not prefixes:
                        raise BuilderError("it's empty")
                    dest_list.extend(
                        prefixes
                    )
                except ARouteServerError as e:
                    errors = True
                    logging.error(
                        "Error while retrieving r_set "
                        "{} for {} IPv{}: {}".format(
                            as_set, client_id, ip_ver, str(e)
                        )
                    )

        if errors:
            raise BuilderError()

    def enrich_config_rpsl(self, client):
        client_rpsl = client["cfg"]["filtering"]["rpsl"]
        if not client_rpsl["enforce_origin_in_as_set"] and \
            not client_rpsl["enforce_prefix_in_as_set"] and \
            not self.cfg_general["filtering"]["rpsl"]["tag_as_set"]:

            return

        client["cfg"]["filtering"]["rpsl"]["as_set_asns"] = []
        client["cfg"]["filtering"]["rpsl"]["as_set_prefixes"] = []

        if client["cfg"]["filtering"]["rpsl"]["as_sets"]:
            as_sets = client["cfg"]["filtering"]["rpsl"]["as_sets"]
        else:
            as_sets = ["AS{}".format(client["as"])]

        errors = False
        try:
            self.enrich_config_rpsl_as_set(
                client["id"],
                as_sets,
                client["cfg"]["filtering"]["rpsl"]["as_set_asns"]
            )
        except ARouteServerError as e:
            errors = True
            logging.error(
                "One or more errors occurred while expanding client's "
                "AS-SETs to obtain the list of allowed origin ASNs; "
                "client ID: {}".format(client["id"])
            )

        try:
            self.enrich_config_rpsl_r_set(
                client["id"],
                as_sets,
                client["cfg"]["filtering"]["rpsl"]["as_set_prefixes"]
            )
        except ARouteServerError as e:
            errors = True
            logging.error(
                "One or more errors occurred while expanding client's "
                "AS-SETs to obtain the list of allowed prefixes; "
                "client ID: {}".format(client["id"])
            )

        if errors:
            raise BuilderError()

    def enrich_config_peeringdb(self, client):
        client_max_prefix = client["cfg"]["filtering"]["max_prefix"]
        if not client_max_prefix["action"]:
            # No max-prefix action given for this client:
            # no needs to know its max-pref limit.
            return

        if client_max_prefix["limit_ipv{}".format(self.ip_ver)]:
            # Client uses a specific limit:
            # no needs to gather info from PeeringDB.
            return

        general_limit_ipv4 = self.cfg_general["filtering"]["max_prefix"]["general_limit_ipv4"]
        general_limit_ipv6 = self.cfg_general["filtering"]["max_prefix"]["general_limit_ipv6"]

        if not client_max_prefix["peering_db"]:
            client_max_prefix["limit_ipv4"] = general_limit_ipv4
            client_max_prefix["limit_ipv6"] = general_limit_ipv6
            return
        
        try:
            net = PeeringDBNet(client["as"])
            client_max_prefix["limit_ipv4"] = net.info_prefixes4 or general_limit_ipv4
            client_max_prefix["limit_ipv6"] = net.info_prefixes6 or general_limit_ipv6

        except PeeringDBNoInfoError:
            # No data found on PeeringDB.
            pass
        except PeeringDBError as e:
            raise BuilderError(
                "An error occurred while retrieving info from PeeringDB "
                "for ASN {}: {}".format(
                    client["as"], str(e) or "error unknown"
                )
            )

    def enrich_config(self):
        errors = False

        # Unique ASNs from clients list.
        clients_asns = {}

        for client in self.cfg_clients.cfg["clients"]:
            # Unique ASNs
            asn = "AS{}".format(client["as"])
            if asn in clients_asns:
                clients_asns[asn] += 1
            else:
                clients_asns[asn] = 1

            # Client's ID
            # Set 'id' as ASx_y where
            # x = ASN
            # y = progressive counter of clients per ASN
            client_id = "{}_{}".format(
                asn, clients_asns[asn]
            )
            client["id"] = client_id

            # RPSL info
            try:
                self.enrich_config_rpsl(client)
            except ARouteServerError as e:
                errors = True
                if str(e):
                    logging.error(str(e))

            # PeerindDB info
            try:
                self.enrich_config_peeringdb(client)
            except ARouteServerError as e:
                errors = True
                if str(e):
                    logging.error(str(e))

        if errors:
            raise BuilderError()

    def render_template(self):
        data = {}
        data["ip_ver"] = self.ip_ver
        data["cfg"] = self.cfg_general
        data["bogons"] = self.cfg_bogons
        data["clients"] = self.cfg_clients

        def current_ipver(ip):
            if self.ip_ver is None:
                return True
            return ipaddr.IPAddress(ip).version == self.ip_ver

        def community_is_set(comm):
            if not comm:
                return False
            if not comm["std"] and not comm["lrg"] and not comm["ext"]:
                return False
            return True

        env = Environment(
            loader=FileSystemLoader(self.template_dir),
            trim_blocks=True,
            lstrip_blocks=True
        )
        env.tests["current_ipver"] = current_ipver
        env.filters["community_is_set"] = community_is_set

        tpl = env.get_template(self.template_name)
        return tpl.render(data)

class BIRDConfigBuilder(ConfigBuilder):

    def validate_flavor_specific_configuration(self):
        if self.ip_ver is None:
            raise BuilderError(
                "An explicit target IP version is needed "
                "to build BIRD configuration. Use the "
                "--ip-ver command line argument to supply one."
            )



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  Ifnot, see <http://www.gnu.org/licenses/>.

from copy import deepcopy
import logging
import os
import sys
import yaml

from ..resources import get_config_dir, get_templates_dir
from ..errors import ConfigError, ARouteServerError


class ConfigParserProgram(object):

    DEFAULT_CFG_DIR = "/etc/arouteserver"
    DEFAULT_CFG_PATH = "{}/arouteserver.yml".format(DEFAULT_CFG_DIR)

    DEFAULT = {
        "cfg_dir": "{}".format(DEFAULT_CFG_DIR),

        "logging_config_file": "log.ini",

        "cfg_general": "general.yml",
        "cfg_clients": "clients.yml",
        "cfg_bogons": "bogons.yml",

        "template_dir": "templates",
        "template_name": "main.j2",

        "cache_dir": "/var/lib/arouteserver",

        "bgpq3_path": "bgpq3",
        "cache_expiry_time": 43200
    }

    def __init__(self):
        self.cfg = deepcopy(self.DEFAULT)

    def load(self, path):
        try:
            with open(path, "r") as f:
                cfg_from_file = yaml.load(f.read())
                if cfg_from_file:
                    self.cfg.update(cfg_from_file)
        except Exception as e:
            logging.error("An error occurred while reading global "
                          "configuration at {}: {}".format(path, str(e)),
                          exc_info=True)
            raise ConfigError()

    def get_cfg_file_path(self, cfg_key):
        assert cfg_key in ("logging_config_file", "cfg_general", "cfg_clients",
                           "cfg_bogons", "template_dir", "cache_dir")

        val = self.cfg[cfg_key]
        if os.path.isabs(val):
            return val
        return os.path.join(self.cfg["cfg_dir"], val)

    def setup(self):

        print("ARouteServer setup")
        print("")

        config_dir = get_config_dir()
        templates_dir = get_templates_dir()

        sys.stdout.write("Where do you want configuration files and templates "
                         "to be stored (default: {}): ".format(
                             self.DEFAULT_CFG_DIR
                         ))
        try:
            dest_dir = raw_input()
        except:
            print("")
            print("Setup aborted")
            return False

        if not dest_dir:
            dest_dir = self.DEFAULT_CFG_DIR
        dest_dir = dest_dir.strip()

        if dest_dir != self.DEFAULT_CFG_DIR:
            print("WARNING: the directory that has been chosen is not the "
                  "default one: use the --cfg command line argument to "
                  "allow the program to find the needed files.")

        sys.stdout.write("Do you confirm you want ARouteServer files to be "
                         "stored at {}: [YES/no] ".format(dest_dir))
        try:
            yes_no = raw_input()
        except:
            print("")
            print("Setup aborted")
            return False

        if not yes_no:
            yes_no = "yes"

        if yes_no.lower() != "yes":
            print("")
            print("Setup aborted")
            return False

        def mk_dir(d):
            sys.stdout.write("Creating {}... ".format(d))
            if os.path.exists(d):
                print("already exists")
            else:
                os.makedirs(d)
                print("OK")

        def cp_file(s, d):
            sys.stdout.write(
                "Copying default {} file into {}... ".format(
                    os.path.basename(s), d
                )
            )

            if os.path.exists(d):
                sys.stdout.write("already exists: do you want to overwrite it [yes/NO]: ")
                try:
                    yes_no = raw_input()
                except:
                    return False

                if not yes_no:
                    yes_no = "no"

                if yes_no.lower() != "yes":
                    print("skipping")
                    return True

            with open(s, "r") as src:
                with open(d, "w") as dst:
                    dst.write(src.read())
            print("OK")
            return True

        mk_dir(dest_dir)
        mk_dir(os.path.join(dest_dir, "templates"))

        def process_dir(s, d):
            for filename in os.listdir(s):
                if os.path.isdir(os.path.join(s, filename)):
                    mk_dir(os.path.join(d, filename))
                    process_dir(os.path.join(s, filename),
                                os.path.join(d, filename))
                else:
                    if not cp_file(os.path.join(s, filename),
                                   os.path.join(d, filename)):
                        print("")
                        print("Setup aborted")
                        return False
            return True

        if not process_dir(config_dir, dest_dir):
            return False
        if not process_dir(templates_dir, os.path.join(dest_dir, "templates")):
            return False

program_config = ConfigParserProgram()



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .base import ConfigParserBase
from .validators import *
from ..errors import ConfigError, ARouteServerError


class ConfigParserGeneral(ConfigParserBase):

    ROOT = "cfg"

    def parse(self):
        """
        Contents of cfg dict is updated/normalized by validators.
        """

        errors = False

        schema = {
            "cfg": {
                "rs_as": ValidatorASN(),
                "router_id": ValidatorIPv4Addr(mandatory=True),
                "prepend_rs_as": ValidatorBool(default=False),                      # Done
                "path_hiding": ValidatorBool(default=True),                         # Done, not tested
                "passive": ValidatorBool(default=True),                             # Done
                "gtsm": ValidatorBool(default=True),
                "add_path": ValidatorBool(default=False),                           # Done
                "filtering": {
                    "next_hop_policy": ValidatorOption("next_hop_policy",
                                                       ("strict", "same-as"),
                                                       default="strict"),           # Done
                    "ipv4_pref_len": ValidatorIPMinMaxLen(4,
                                                          default={"min": 8,
                                                                   "max": 24}),     # Done
                    "ipv6_pref_len": ValidatorIPMinMaxLen(6,
                                                          default={"min": 12,
                                                                   "max": 48}),     # Done, not tested
                    "global_black_list_pref": ValidatorListOf(
                        ValidatorPrefixListEntry, mandatory=False,
                    ),                                                              # Done
                    "max_as_path_len": ValidatorMaxASPathLen(default=32),           # Done
                    "reject_invalid_as_in_as_path": ValidatorBool(default=True),    # Done
                    "rpsl": {
                        "enforce_origin_in_as_set": ValidatorBool(default=True),    # Done
                        "enforce_prefix_in_as_set": ValidatorBool(default=True),    # Done
                        "tag_as_set": ValidatorBool(default=True)                   # Done
                    },
                    "rpki": {
                        "enabled": ValidatorBool(default=False),                    # MISSING
                        "data_source": ValidatorOption("data_source",
                                                       ("rtrsub", "rtrlib"),
                                                       mandatory=False),            # MISSING
                        "reject_invalid": ValidatorBool(mandatory=False),           # MISSING
                    },
                    "max_prefix": {                                                 # MISSING
                        "peering_db": ValidatorBool(default=True),
                        "general_limit_ipv4": ValidatorUInt(default=170000),
                        "general_limit_ipv6": ValidatorUInt(default=12000),
                        "action": ValidatorOption(
                            "action",
                            ("shutdown", "restart", "block", "warning"),
                            mandatory=False,
                            default="shutdown"
                        )
                    },
                },
                "blackhole_filtering": {                                        # Done
                    "announce_to_client": ValidatorBool(                        # Done
                        mandatory=True, default=True
                    ),
                    "policy_ipv4": ValidatorOption(                             # Done
                        "policy_ipv4",
                        ("propagate-unchanged", "rewrite-next-hop"),
                        mandatory=False),
                    "policy_ipv6": ValidatorOption(                             # Done, not tested
                        "policy_ipv6",
                        ("propagate-unchanged", "rewrite-next-hop"),
                        mandatory=False),
                    "rewrite_next_hop_ipv4": ValidatorIPv4Addr(mandatory=False),# Done
                    "rewrite_next_hop_ipv6": ValidatorIPv6Addr(mandatory=False),# Done, not tested
                },
                "control_communities": ValidatorBool(default=True),                 # Done
                "communities": {
                }
            }
        }

        if "rs_as" in self.cfg["cfg"]:
            rs_as_macro = self.cfg["cfg"]["rs_as"]
        else:
            rs_as_macro = None

        # Add communities with no fixed values to the schema.
        for comm in ["origin_present_in_as_set", "origin_not_present_in_as_set",
                     "prefix_present_in_as_set", "prefix_not_present_in_as_set",
                     "roa_valid", "roa_invalid", "roa_unknown",
                     "blackholing",
                     "do_not_announce_to_any",
                     "prepend_once_to_any", "prepend_twice_to_any",
                     "prepend_thrice_to_any"]:


            schema["cfg"]["communities"][comm] = {
                "std": ValidatorCommunityStd(rs_as_macro, mandatory=False),
                "lrg": ValidatorCommunityLrg(rs_as_macro, mandatory=False),
                "ext": ValidatorCommunityExt(rs_as_macro, mandatory=False),
            }

        # Add communities with fixed values to the schema.
        for comm, fixed_values in [
            ("do_not_announce_to_peer",
                ("0:peer_as",
                 "rs_as:0:peer_as",
                 ["rt:0:peer_as", "ro:0:peer_as"])
            ),
            ("announce_to_peer",
                ("rs_as:peer_as",
                 "rs_as:rs_as:peer_as",
                 ["rt:rs_as:peer_as", "ro:rs_as:peer_as"])
            )]:
            schema["cfg"]["communities"][comm] = {
                "std": ValidatorCommunityStd(rs_as_macro, mandatory=False,
                                            fixed_values=fixed_values[0],
                                            allow_peer_as_macro=True),
                "lrg": ValidatorCommunityLrg(rs_as_macro, mandatory=False,
                                            fixed_values=fixed_values[1],
                                            allow_peer_as_macro=True),
                "ext": ValidatorCommunityExt(rs_as_macro, mandatory=False,
                                            fixed_values=fixed_values[2],
                                            allow_peer_as_macro=True)
            }

        try:
            ConfigParserBase.validate(schema, self.cfg)
        except ARouteServerError as e:
            errors = True
            if str(e):
                logging.error(str(e))

        # If blackhole filtering policy = "rewrite-next-hop", then
        # blackhole next-hops must be provided.
        for ip_ver in (4, 6):
            bh = self.cfg["cfg"]["blackhole_filtering"]
            if not bh:
                continue
            policy = bh["policy_ipv{}".format(ip_ver)]
            if policy == "rewrite-next-hop":
                if not bh["rewrite_next_hop_ipv{}".format(ip_ver)]:
                    errors = True
                    logging.error(
                        "Since blackhole_filtering.policy_ipv{v} is "
                        "'rewrite_next_hop', an IPv{v} address must "
                        "be provided in "
                        "'rewrite_next_hop_ipv{v}'.".format(
                            v=ip_ver
                        )
                    )

        # Duplicate communities?
        communities = self.cfg["cfg"]["communities"]
        unique_communities = []
        for comm_tag in communities:
            comm = communities[comm_tag]
            for fmt in ("std", "lrg", "ext"):
                if comm[fmt]:
                    if comm[fmt] in unique_communities:
                        errors = True
                        logging.error(
                            "The '{}.{}' community's value ({}) "
                            "has already been used for another "
                            "community.".format(comm_tag, fmt, comm[fmt])
                        )
                    else:
                        unique_communities.append(comm[fmt])

        if errors:
            raise ConfigError()



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .base import TagASSetScenario
from ...bird import BIRDInstanceIPv6

class TagASSetScenario_BIRDIPv6(TagASSetScenario):
    __test__ = True

    SHORT_DESCR = "Live test, BIRD, tag prefix/origin in AS-SET, IPv6"
    RS_INSTANCE_CLASS = BIRDInstanceIPv6
    CLIENT_INSTANCE_CLASS = BIRDInstanceIPv6
    IP_VER = 6

    DATA = {
        "rs_IPAddress":                     "2001:db8:1:1::2",
        "AS1_1_IPAddress":                  "2001:db8:1:1::11",
        "AS2_1_IPAddress":                  "2001:db8:1:1::21",
        "AS4_1_IPAddress":                  "2001:db8:1:1::41",
        "AS5_1_IPAddress":                  "2001:db8:1:1::51",

        "AS1_allowed_prefixes":             "2a01:0::/32",

        "AS2_allowed_prefixes":             "2a02:0::/32",
        "AS2_pref_ok_origin_ok1":           "2a02:0:1::/48",
        "AS2_pref_ko_origin_ok1":           "2a02:1:0::/48",
        "AS3_pref_ok_origin_ko1":           "2a02:0:2::/48",
        "AS3_pref_ko_origin_ko1":           "2a03:0:1::/48",

        "AS4_allowed_prefixes":             "2a04:0::/32",
        "AS4_pref_ok_origin_ok1":           "2a04:0:1::/48",
        "AS4_pref_ko_origin_ok1":           "2a04:1:0::/48",
        "AS3_pref_ok_origin_ko2":           "2a04:0:2::/48",

        "AS5_allowed_prefixes":             "2a05:0::/32",
        "AS5_pref_ok_origin_ok1":           "2a05:0:1::/48",
        "AS5_pref_ko_origin_ok1":           "2a05:1:0::/48",
        "AS3_pref_ok_origin_ko3":           "2a05:0:2::/48",
    }



import os
from os.path import abspath, dirname, join
from setuptools import setup, find_packages

"""
New release procedure

- nosetests -vs tests/

- edit pierky/arouteserver/version.py

- edit CHANGES.rst

- verify RST syntax is ok
    python setup.py --long-description | rst2html.py --strict

- build and verify docs
    cd docs ; make html ; python -m SimpleHTTPServer 8000 ; cd ..

- new files to be added to MANIFEST.in?

- python setup.py sdist

- twine upload dist/*

- git push

- edit new release on GitHub
"""

__version__ = None

# Allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# Get proper long description for package
current_dir = dirname(abspath(__file__))
description = open(join(current_dir, "README.rst")).read()
changes = open(join(current_dir, "CHANGES.rst")).read()
long_description = '\n\n'.join([description, changes])
exec(open(join(current_dir, "pierky/arouteserver/version.py")).read())

install_requires = []
with open("requirements.txt", "r") as f:
    for line in f.read().split("\n"):
        if line:
            install_requires.append(line)
print(install_requires)

# Get the long description from README.md
setup(
    name="arouteserver",
    version=__version__,

    packages=["pierky", "pierky.arouteserver"],
    namespace_packages=["pierky"],
    data_files=[
        ("/arouteserver/config.d", [os.path.join("config.d", f) for f in os.listdir("config.d")]),
        ("/arouteserver/templates/bird", [os.path.join("templates/bird", f) for f in os.listdir("templates/bird")]),
    ],
    include_package_data=True,
    
    license="GPLv3",
    description="A Python tool to automatically build (and test) configurations for BGP route servers.",
    long_description=long_description,
    url="https://github.com/pierky/arouteserver",
    download_url="https://github.com/pierky/arouteserver",

    author="Pier Carlo Chiodi",
    author_email="pierky@pierky.com",
    maintainer="Pier Carlo Chiodi",
    maintainer_email="pierky@pierky.com",

    install_requires=install_requires,
    tests_require=[
        "nose",
        "mock",
    ],
    test_suite="nose.collector",

    scripts=["scripts/arouteserver"],

    keywords=['BGP', 'Route server', 'BIRD', 'IP Routing'],

    classifiers=[
        "Development Status :: 4 - Beta",

        "Environment :: Console",

        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",

        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",

        "Operating System :: POSIX",
        "Operating System :: Unix",

        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",

        "Topic :: Internet :: WWW/HTTP",
        "Topic :: System :: Networking",
    ],
)
